/*------------------------------------------------------------
 * BLH2XYZ.c - Coordinate Conversion.
 *------------------------------------------------------------*/

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>

/*------------------------------------------------------------
 * 萔E\̂̒`
 *------------------------------------------------------------*/
/* _^ */
typedef	int	bool;
#define	TRUE	1
#define	FALSE	0

/* WGS-84萔 */
#define	PI		3.1415926535898		/* ~iIS-GPS-200j */
#define	C		2.99792458e8		/* [m/s] */
#define	MUe		3.986005e14			/* nd͒萔[m^3/s^2] */
#define	dOMEGAe	7.2921151467e-5		/* n]px[rad/s] */
#define	Re		6378137.0			/* na[m] */
#define	Fe		(1.0/298.257223563)	/* n̝G */

/* px̕ϊ */
#define	rad_to_deg(rad)		((rad)/PI*180.0)
#define	deg_to_rad(deg)		((deg)/180.0*PI)
#define	rad_to_sc(rad)		((rad)/PI)
#define	sc_to_rad(sc)		((sc)*PI)

/* 舵s̑傫 */
#define	MAX_N				16		/* ϑq̏ */
#define	MAX_M				4		/* m̍ő吔 */
#define	MAX_PRN				32		/* qԍ̏ */

/*  */
#define	SECONDS_DAY			(3600L*24L)
#define	SECONDS_WEEK		(3600L*24L*7L)

/* \\ */
typedef struct {
	int		week;		/* Tԍ */
	double	sec;		/* T߂̌oߎ[s] */
}	wtime;

/* W\\ */
typedef struct {
	double	x;			/* XW[m] */
	double	y;			/* YW[m] */
	double	z;			/* ZW[m] */
}	posxyz;
#define	SQ(x)			((x)*(x))
#define	DIST(a,b)		sqrt(SQ(a.x-b.x)+SQ(a.y-b.y)+SQ(a.z-b.z))

/* oܓx\\ */
typedef struct {
	double	lat;		/* ܓx[rad] */
	double	lon;		/* ox[rad] */
	double	hgt;		/* xiȉ~̍j[m] */
}	posblh;

/* ENUW\\ */
typedef struct {
	double	e;			/* East[m] */
	double	n;			/* North[m] */
	double	u;			/* Up[m] */
}	posenu;

/*------------------------------------------------------------
 * t@C̓Ǎ
 *------------------------------------------------------------*/
/* sobt@ */
#define	LINEBUF_LEN			256
static char	linebuf[LINEBUF_LEN],fieldbuf[LINEBUF_LEN];
static int	linepos			=0;

/* t@Csobt@ɓǂݍ */
static bool read_line(FILE *fp)
{
	/* sobt@NA */
	linepos		=0;
	linebuf[0]	='\0';

	/* t@Cǂݍ */
	return (fgets(linebuf,LINEBUF_LEN,fp)!=NULL);
}

/* sobt@當oiwA܂CSV`j */
static char *get_field(int width)
{
	int		i;
	bool	quotef=TRUE;

	/* width==0ȂCSV` */
	if (width<1) {
		width	=LINEBUF_LEN-1;
		quotef	=FALSE;
	}

	/* w肳ꂽ邢̓R}܂łǂݎ */
	for(i=0;i<width && linebuf[linepos+i]!='\0';i++) {
		if (linebuf[linepos+i]=='\n') break;
		if (!quotef && linebuf[linepos+i]==',') {
			linepos++;
			break;
		}
		if (linebuf[linepos+i]=='\"') {
			quotef=!quotef;
			continue;
		}
		fieldbuf[i]=linebuf[linepos+i];
	}
	linepos		+=i;		/* ̈ʒu */
	fieldbuf[i]	='\0';

	return fieldbuf;
}

/*------------------------------------------------------------
 * Wϊ
 *------------------------------------------------------------*/

/*------------------------------------------------------------
 * xyz_to_blh() - Woܓxւ̕ϊ
 *------------------------------------------------------------
 *  posblh xyz_to_blh(pos); oܓx
 *    posxyz pos; Wl
 *------------------------------------------------------------*/
posblh xyz_to_blh(posxyz pos)
{
	double	a,b,e,f,n,h,p,t,sint,cost;
	posblh	blh={0.0,0.0,-Re};

	/* _̏ꍇ */
	if (pos.x==0.0 && pos.y==0.0 && pos.z==0.0) return blh;

	/* ȉ~̂̃p[^ */
	f	=Fe;				/* G */
	a	=Re;				/* a */
	b	=a*(1.0-f);			/* Za */
	e	=sqrt(f*(2.0-f));	/* S */

	/* Wϊ̂߂̃p[^ */
	h	=a*a-b*b;
	p	=sqrt(pos.x*pos.x+pos.y*pos.y);
	t	=atan2(pos.z*a,p*b);
	sint=sin(t);
	cost=cos(t);

	/* oܓxւ̕ϊ */
	blh.lat =atan2(pos.z+h/b*sint*sint*sint,p-h/a*cost*cost*cost);
	n		=a/sqrt(1.0-e*e*sin(blh.lat)*sin(blh.lat));
	blh.lon	=atan2(pos.y,pos.x);
	blh.hgt	=(p/cos(blh.lat))-n;

	return blh;
}

/*------------------------------------------------------------
 * blh_to_xyz() - oܓx璼Wւ̕ϊ
 *------------------------------------------------------------
 *  posxyz blh_to_xyz(blh); Wl
 *    posblh blh; oܓx
 *------------------------------------------------------------*/
posxyz blh_to_xyz(posblh blh)
{
	double	a,b,e,f,n;
	posxyz	pos;

	/* ȉ~̂̃p[^ */
	f	=Fe;				/* G */
	a	=Re;				/* a */
	b	=a*(1.0-f);			/* Za */
	e	=sqrt(f*(2.0-f));	/* S */

	/* Wnւ̕ϊ */
	n		=a/sqrt(1.0-e*e*sin(blh.lat)*sin(blh.lat));
	pos.x	=(n+blh.hgt)*cos(blh.lat)*cos(blh.lon);
	pos.y	=(n+blh.hgt)*cos(blh.lat)*sin(blh.lon);
	pos.z	=(n*(1.0-e*e)+blh.hgt)*sin(blh.lat);

	return pos;
}

/*------------------------------------------------------------
 * xyz_to_enu() - WENUWւ̕ϊ
 *------------------------------------------------------------
 *  posenu xyz_to_enu(pos,base); ENUWl
 *    posxyz pos;  Wl
 *    posxyz base; ʒu
 *------------------------------------------------------------*/
posenu xyz_to_enu(posxyz pos,posxyz base)
{
	double	s1,c1,s2,c2;
	posblh	blh;
	posenu	enu;

	/* ʒȗΈʒu */
	pos.x	-=base.x;
	pos.y	-=base.y;
	pos.z	-=base.z;

	/* ʒǔoܓx */
	blh	=xyz_to_blh(base);
	s1	=sin(blh.lon);
	c1	=cos(blh.lon);
	s2	=sin(blh.lat);
	c2	=cos(blh.lat);

	/* Έʒu]ENUWɕϊ */
	enu.e	=-pos.x*s1+pos.y*c1;
	enu.n	=-pos.x*c1*s2-pos.y*s1*s2+pos.z*c2;
	enu.u	=pos.x*c1*c2+pos.y*s1*c2+pos.z*s2;

	return enu;
}

/*------------------------------------------------------------
 * enu_to_xyz() - ENUW璼Wւ̕ϊ
 *------------------------------------------------------------
 *  posxyz enu_to_xyz(enu,base); Wl
 *    posenu enu;  ENUWl
 *    posxyz base; ʒu
 *------------------------------------------------------------*/
posxyz enu_to_xyz(posenu enu,posxyz base)
{
	double	s1,c1,s2,c2;
	posblh	blh;
	posxyz	pos;

	/* ʒǔoܓx */
	blh	=xyz_to_blh(base);
	s1	=sin(blh.lon);
	c1	=cos(blh.lon);
	s2	=sin(blh.lat);
	c2	=cos(blh.lat);

	/* ENUW]đΈʒuɕϊ */
	pos.x	=-enu.e*s1-enu.n*c1*s2+enu.u*c1*c2;
	pos.y	=enu.e*c1-enu.n*s1*s2+enu.u*s1*c2;
	pos.z	=enu.n*c2+enu.u*s2;

	/* ʒuɉ */
	pos.x	+=base.x;
	pos.y	+=base.y;
	pos.z	+=base.z;

	return pos;
}

/*------------------------------------------------------------
 * elevation() - p߂
 *------------------------------------------------------------
 *  double elevation(sat,usr); p[rad]
 *    posxyz sat; q̈ʒu
 *    posxyz usr; [Uʒu
 *------------------------------------------------------------*/
double elevation(posxyz sat,posxyz usr)
{
	posenu	enu;

	/* ENUWɕϊċp߂ */
	enu	=xyz_to_enu(sat,usr);
	return atan2(enu.u,sqrt(enu.e*enu.e+enu.n*enu.n));
}

/*------------------------------------------------------------
 * azimuth() - ʊp߂
 *------------------------------------------------------------
 *  double azimuth(sat,usr); ʊp[rad]
 *    posxyz sat; q̈ʒu
 *    posxyz usr; [Uʒu
 *------------------------------------------------------------*/
double azimuth(posxyz sat,posxyz usr)
{
	posenu	enu;

	/* ENUWɕϊĕʊp߂ */
	enu	=xyz_to_enu(sat,usr);
	return atan2(enu.e,enu.n);
}

/*------------------------------------------------------------
 * main() - C
 *------------------------------------------------------------*/
void main(int argc,char **argv)
{
	double	sec;
	posxyz	pos,rel_pos;
	posblh	blh;

	/* R}hC`FbN */
	if (argc!=1) {
		printf("blh2xyz - convert LAT/LON/HGT into XYZ.\n");
		printf("\n");
		printf("usage: blh2xyz <in >out\n");
		printf("\n");
		exit(0);
	}

	/* f[^t@Cǂݍ */
	while(read_line(stdin)) {
		sec		=atof(get_field(0));			/*  */
		blh.lat	=deg_to_rad(atof(get_field(0)));/* ܓx[deg] */
		blh.lon	=deg_to_rad(atof(get_field(0)));/* ox[deg] */
		blh.hgt	=atof(get_field(0));			/* x[m] */

		/* Wϊďo͂ */
		printf("%.3f,%.4f,%.4f,%.4f,%s\n",
			sec,								/* [s] */
			blh_to_xyz(blh).x,					/* XW[m] */
			blh_to_xyz(blh).y,					/* YW[m] */
			blh_to_xyz(blh).z,					/* ZW[m] */
			get_field(LINEBUF_LEN));			/* c */
	}

	exit(0);
}

