/*------------------------------------------------------------
 * MASKSATS.c - Mask Angle versus the Number of Satellites.
 *------------------------------------------------------------*/

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>

/*------------------------------------------------------------
 * 萔E\̂̒`
 *------------------------------------------------------------*/
/* _^ */
typedef	int	bool;
#define	TRUE	1
#define	FALSE	0

/* WGS-84萔 */
#define	PI		3.1415926535898		/* ~iIS-GPS-200j */
#define	C		2.99792458e8		/* [m/s] */
#define	MUe		3.986005e14			/* nd͒萔[m^3/s^2] */
#define	dOMEGAe	7.2921151467e-5		/* n]px[rad/s] */
#define	Re		6378137.0			/* na[m] */
#define	Fe		(1.0/298.257223563)	/* n̝G */

/* px̕ϊ */
#define	rad_to_deg(rad)		((rad)/PI*180.0)
#define	deg_to_rad(deg)		((deg)/180.0*PI)
#define	rad_to_sc(rad)		((rad)/PI)
#define	sc_to_rad(sc)		((sc)*PI)

/* 舵s̑傫 */
#define	MAX_N				16		/* ϑq̏ */
#define	MAX_M				4		/* m̍ő吔 */
#define	MAX_PRN				32		/* qԍ̏ */

/*  */
#define	SECONDS_DAY			(3600L*24L)
#define	SECONDS_WEEK		(3600L*24L*7L)

/* \\ */
typedef struct {
	int		week;		/* Tԍ */
	double	sec;		/* T߂̌oߎ[s] */
}	wtime;

/* W\\ */
typedef struct {
	double	x;			/* XW[m] */
	double	y;			/* YW[m] */
	double	z;			/* ZW[m] */
}	posxyz;
#define	SQ(x)			((x)*(x))
#define	DIST(a,b)		sqrt(SQ(a.x-b.x)+SQ(a.y-b.y)+SQ(a.z-b.z))

/* oܓx\\ */
typedef struct {
	double	lat;		/* ܓx[rad] */
	double	lon;		/* ox[rad] */
	double	hgt;		/* xiȉ~̍j[m] */
}	posblh;

/* ENUW\\ */
typedef struct {
	double	e;			/* East[m] */
	double	n;			/* North[m] */
	double	u;			/* Up[m] */
}	posenu;

/*------------------------------------------------------------
 * s̃p[^
 *------------------------------------------------------------*/
/* p}XN͈̔ */
#define	MASK_START			0.0
#define	MASK_STOP			50.0
#define	MASK_STEP			1.0

/*------------------------------------------------------------
 * t@C̓Ǎ
 *------------------------------------------------------------*/
/* sobt@ */
#define	LINEBUF_LEN			256
static char	linebuf[LINEBUF_LEN],fieldbuf[LINEBUF_LEN];
static int	linepos			=0;

/* t@Csobt@ɓǂݍ */
static bool read_line(FILE *fp)
{
	/* sobt@NA */
	linepos		=0;
	linebuf[0]	='\0';

	/* t@Cǂݍ */
	return (fgets(linebuf,LINEBUF_LEN,fp)!=NULL);
}

/* sobt@當oiwA܂CSV`j */
static char *get_field(int width)
{
	int		i;
	bool	quotef=TRUE;

	/* width==0ȂCSV` */
	if (width<1) {
		width	=LINEBUF_LEN-1;
		quotef	=FALSE;
	}

	/* w肳ꂽ邢̓R}܂łǂݎ */
	for(i=0;i<width && linebuf[linepos+i]!='\0';i++) {
		if (linebuf[linepos+i]=='\n') break;
		if (!quotef && linebuf[linepos+i]==',') {
			linepos++;
			break;
		}
		if (linebuf[linepos+i]=='\"') {
			quotef=!quotef;
			continue;
		}
		fieldbuf[i]=linebuf[linepos+i];
	}
	linepos		+=i;		/* ̈ʒu */
	fieldbuf[i]	='\0';

	return fieldbuf;
}

/*------------------------------------------------------------
 * Wϊ
 *------------------------------------------------------------*/

/*------------------------------------------------------------
 * xyz_to_blh() - Woܓxւ̕ϊ
 *------------------------------------------------------------
 *  posblh xyz_to_blh(pos); oܓx
 *    posxyz pos; Wl
 *------------------------------------------------------------*/
posblh xyz_to_blh(posxyz pos)
{
	double	a,b,e,f,n,h,p,t,sint,cost;
	posblh	blh={0.0,0.0,-Re};

	/* _̏ꍇ */
	if (pos.x==0.0 && pos.y==0.0 && pos.z==0.0) return blh;

	/* ȉ~̂̃p[^ */
	f	=Fe;				/* G */
	a	=Re;				/* a */
	b	=a*(1.0-f);			/* Za */
	e	=sqrt(f*(2.0-f));	/* S */

	/* Wϊ̂߂̃p[^ */
	h	=a*a-b*b;
	p	=sqrt(pos.x*pos.x+pos.y*pos.y);
	t	=atan2(pos.z*a,p*b);
	sint=sin(t);
	cost=cos(t);

	/* oܓxւ̕ϊ */
	blh.lat =atan2(pos.z+h/b*sint*sint*sint,p-h/a*cost*cost*cost);
	n		=a/sqrt(1.0-e*e*sin(blh.lat)*sin(blh.lat));
	blh.lon	=atan2(pos.y,pos.x);
	blh.hgt	=(p/cos(blh.lat))-n;

	return blh;
}

/*------------------------------------------------------------
 * blh_to_xyz() - oܓx璼Wւ̕ϊ
 *------------------------------------------------------------
 *  posxyz blh_to_xyz(blh); Wl
 *    posblh blh; oܓx
 *------------------------------------------------------------*/
posxyz blh_to_xyz(posblh blh)
{
	double	a,b,e,f,n;
	posxyz	pos;

	/* ȉ~̂̃p[^ */
	f	=Fe;				/* G */
	a	=Re;				/* a */
	b	=a*(1.0-f);			/* Za */
	e	=sqrt(f*(2.0-f));	/* S */

	/* Wnւ̕ϊ */
	n		=a/sqrt(1.0-e*e*sin(blh.lat)*sin(blh.lat));
	pos.x	=(n+blh.hgt)*cos(blh.lat)*cos(blh.lon);
	pos.y	=(n+blh.hgt)*cos(blh.lat)*sin(blh.lon);
	pos.z	=(n*(1.0-e*e)+blh.hgt)*sin(blh.lat);

	return pos;
}

/*------------------------------------------------------------
 * xyz_to_enu() - WENUWւ̕ϊ
 *------------------------------------------------------------
 *  posenu xyz_to_enu(pos,base); ENUWl
 *    posxyz pos;  Wl
 *    posxyz base; ʒu
 *------------------------------------------------------------*/
posenu xyz_to_enu(posxyz pos,posxyz base)
{
	double	s1,c1,s2,c2;
	posblh	blh;
	posenu	enu;

	/* ʒȗΈʒu */
	pos.x	-=base.x;
	pos.y	-=base.y;
	pos.z	-=base.z;

	/* ʒǔoܓx */
	blh	=xyz_to_blh(base);
	s1	=sin(blh.lon);
	c1	=cos(blh.lon);
	s2	=sin(blh.lat);
	c2	=cos(blh.lat);

	/* Έʒu]ENUWɕϊ */
	enu.e	=-pos.x*s1+pos.y*c1;
	enu.n	=-pos.x*c1*s2-pos.y*s1*s2+pos.z*c2;
	enu.u	=pos.x*c1*c2+pos.y*s1*c2+pos.z*s2;

	return enu;
}

/*------------------------------------------------------------
 * enu_to_xyz() - ENUW璼Wւ̕ϊ
 *------------------------------------------------------------
 *  posxyz enu_to_xyz(enu,base); Wl
 *    posenu enu;  ENUWl
 *    posxyz base; ʒu
 *------------------------------------------------------------*/
posxyz enu_to_xyz(posenu enu,posxyz base)
{
	double	s1,c1,s2,c2;
	posblh	blh;
	posxyz	pos;

	/* ʒǔoܓx */
	blh	=xyz_to_blh(base);
	s1	=sin(blh.lon);
	c1	=cos(blh.lon);
	s2	=sin(blh.lat);
	c2	=cos(blh.lat);

	/* ENUW]đΈʒuɕϊ */
	pos.x	=-enu.e*s1-enu.n*c1*s2+enu.u*c1*c2;
	pos.y	=enu.e*c1-enu.n*s1*s2+enu.u*s1*c2;
	pos.z	=enu.n*c2+enu.u*s2;

	/* ʒuɉ */
	pos.x	+=base.x;
	pos.y	+=base.y;
	pos.z	+=base.z;

	return pos;
}

/*------------------------------------------------------------
 * elevation() - p߂
 *------------------------------------------------------------
 *  double elevation(sat,usr); p[rad]
 *    posxyz sat; q̈ʒu
 *    posxyz usr; [Uʒu
 *------------------------------------------------------------*/
double elevation(posxyz sat,posxyz usr)
{
	posenu	enu;

	/* ENUWɕϊċp߂ */
	enu	=xyz_to_enu(sat,usr);
	return atan2(enu.u,sqrt(enu.e*enu.e+enu.n*enu.n));
}

/*------------------------------------------------------------
 * azimuth() - ʊp߂
 *------------------------------------------------------------
 *  double azimuth(sat,usr); ʊp[rad]
 *    posxyz sat; q̈ʒu
 *    posxyz usr; [Uʒu
 *------------------------------------------------------------*/
double azimuth(posxyz sat,posxyz usr)
{
	posenu	enu;

	/* ENUWɕϊĕʊp߂ */
	enu	=xyz_to_enu(sat,usr);
	return atan2(enu.e,enu.n);
}

/*------------------------------------------------------------
 * ̕ϊ
 *------------------------------------------------------------*/
/* J_l̊JnN */
#define	TIME_T_BASE_YEAR	1970

/* 1980N1600:00:00̃J_l */
#define	TIME_T_ORIGIN		315964800L

/* mktime()֐GMTŁigmtime()֐ɑΉj */
static time_t mktime2(struct tm *tm)
{
	int		i;
	long	days=0L;
	static int	days_month[]={
		31,28,31,30,31,30,31,31,30,31,30,31
	};

	/* oߓ𓾂 */
	for(i=TIME_T_BASE_YEAR;i<tm->tm_year+1900;i++) {
		days+=(i%4==0)?366:365;
	}
	for(i=1;i<tm->tm_mon+1;i++) {
		days+=days_month[i-1];
		if (i==2 && tm->tm_year%4==0) days++;
	}
	days+=tm->tm_mday-1;

	/* J_lԂ */
	return ((days*24+tm->tm_hour)*60+tm->tm_min)*60+tm->tm_sec;
}

/*------------------------------------------------------------
 * wtime_to_date() - TԍEbւ̕ϊ
 *------------------------------------------------------------
 *  struct tm wtime_to_date(wt); ւ̕ϊ
 *    wtime wt; TԍEb
 *------------------------------------------------------------*/
struct tm wtime_to_date(wtime wt)
{
	time_t	t;

	/* ̌oߎԂāCJ_l𓾂 */
	t	=(long)wt.week*SECONDS_WEEK+TIME_T_ORIGIN
			+(long)((wt.sec>0.0)?wt.sec+0.5:wt.sec-0.5);

	return *gmtime(&t);
}

/*------------------------------------------------------------
 * date_to_wtime() - TԍEbւ̕ϊ
 *------------------------------------------------------------
 *  wtime date_to_wtime(tmbuf); TԍEbւ̕ϊ
 *    struct tm tmbuf; w
 *------------------------------------------------------------*/
wtime date_to_wtime(struct tm tmbuf)
{
	time_t	t;
	wtime	wt;

	/* w肳ꂽ̃J_l */
	t	=mktime2(&tmbuf);

	/* 1TԂ̕bŊƗ] */
	wt.week	=(t-TIME_T_ORIGIN)/SECONDS_WEEK;
	wt.sec	=(t-TIME_T_ORIGIN)%SECONDS_WEEK;

	return wt;
}

/*------------------------------------------------------------
 * q@bZ[W̏
 *------------------------------------------------------------*/
/* RINEXt@C̏ */
#define	RINEX_POS_COMMENT		60
#define	RINEX_NAV_LINES			8
#define	RINEX_NAV_FIELDS_LINE	4

/* LGtFX̍ő吔 */
#define	MAX_EPHMS				20

/* GtFX̗L[h] */
#define	EPHEMERIS_EXPIRE		2.0

/* GtFXi[邽߂̍\ */
typedef	struct {
	int		week;		/* Tԍ */
	double	data[RINEX_NAV_LINES*RINEX_NAV_FIELDS_LINE];
}	ephm_info;

/* p[^ԍ` */
enum ephm_para {
	EPHM_TOC,	EPHM_AF0,	EPHM_AF1,	EPHM_AF2,	/* line 1 */
	EPHM_IODE,	EPHM_Crs,	EPHM_d_n,	EPHM_M0,	/* line 2 */
	EPHM_Cuc,	EPHM_e,		EPHM_Cus,	EPHM_sqrtA,	/* line 3 */
	EPHM_TOE,	EPHM_Cic,	EPHM_OMEGA0,EPHM_Cis,	/* line 4 */
	EPHM_i0,	EPHM_Crc,	EPHM_omega,	EPHM_dOmega,/* line 5 */
	EPHM_di,	EPHM_CAonL2,EPHM_WEEK,	EPHM_L2P,	/* line 6 */
	EPHM_acc,	EPHM_health,EPHM_TGD,	EPHM_IODC,	/* line 7 */
	EPHM_TOT,	EPHM_FIT							/* line 8 */
};

/* GtFXێz */
static ephm_info	ephm_buf[MAX_PRN][MAX_EPHMS];
static int			ephm_count[MAX_PRN];
static int			current_ephm[MAX_PRN];
static int			current_week=-1;

/* [b̏ */
static int			leap_sec	=0;

/* dw␳ */
#define	IONO_PARAMETERS			4
static double	iono_alpha[IONO_PARAMETERS];
static double	iono_beta[IONO_PARAMETERS];

/* 񁨎̕ϊ */
static double atof2(char *str)
{
	char	*p;

	/* 'D''E'ɕϊ */
	for(p=str;*p!='\0';p++) {
		if (*p=='D' || *p=='d') *p='E';
	}

	/* ɕϊĕԂ */
	return atof(str);
}

/* Rg𒲂ׂ */
static bool is_comment(char *str)
{
	return (strncmp(linebuf+RINEX_POS_COMMENT,str,strlen(str))==0);
}

/*------------------------------------------------------------
 * read_RINEX_NAV() - RINEXq@t@Cǂݍ
 *------------------------------------------------------------
 *  void read_RINEX_NAV(fp);
 *    FILE *fp; ǂݍރt@C
 *------------------------------------------------------------*/
void read_RINEX_NAV(FILE *fp)
{
	int		i,j,n,prn,line;
	bool	noerr=FALSE;
	double	d,t;
	wtime	wt;
	struct tm	tmbuf;
	ephm_info	info;

	/*  */
	if (current_week<0) {
		for(i=0;i<IONO_PARAMETERS;i++) {
			iono_alpha[i]	=0.0;
			iono_beta[i]	=0.0;
		}
		for(i=0;i<MAX_PRN;i++) {
			ephm_count[i]	=0;
		}
	}

	/* wb_ */
	while(read_line(fp)) {
		if (is_comment("ION ALPHA")) {
			iono_alpha[0]	=atof2(get_field(14));
			iono_alpha[1]	=atof2(get_field(12));
			iono_alpha[2]	=atof2(get_field(12));
			iono_alpha[3]	=atof2(get_field(12));
		} else if (is_comment("ION BETA")) {
			iono_beta[0]	=atof2(get_field(14));
			iono_beta[1]	=atof2(get_field(12));
			iono_beta[2]	=atof2(get_field(12));
			iono_beta[3]	=atof2(get_field(12));
		} else if (is_comment("LEAP SECONDS")) {
			leap_sec		=atoi(get_field(6));
		} else if (is_comment("END OF HEADER")) break;
	}

	/* {̂ǂ */
	fprintf(stderr,"Reading RINEX NAV... ");
	while(read_line(fp)) {
		n=0;
		for(line=0;line<RINEX_NAV_LINES;line++) {
			/* ŏ̃f[^ǂݍ */
			if (line==0) {
				/* ŏ̍s */
				prn				=atoi(get_field(2));
				tmbuf.tm_year	=atoi(get_field(3));
				if (tmbuf.tm_year<80) tmbuf.tm_year+=100;
				tmbuf.tm_mon	=atoi(get_field(3))-1;
				tmbuf.tm_mday	=atoi(get_field(3));
				tmbuf.tm_hour	=atoi(get_field(3));
				tmbuf.tm_min	=atoi(get_field(3));
				tmbuf.tm_sec	=0;
				wt				=date_to_wtime(tmbuf);
				wt.sec			+=atof(get_field(5));
				info.week		=wt.week;
				d				=wt.sec;
			} else {
				/* 2sڈȍ~ */
				if (!read_line(fp)) goto ERROR;
				d	=atof2(get_field(22));
			}
			info.data[n++]=d;

			/* c̃f[^ */
			for(i=1;i<RINEX_NAV_FIELDS_LINE;i++) {
				d	=atof2(get_field(19));
				info.data[n++]=d;
			}
		}
		if (prn<1) continue;

		/* Tԍ𑵂 */
		t	=(info.week-info.data[EPHM_WEEK])*SECONDS_WEEK;
		info.week			=info.data[EPHM_WEEK];
		info.data[EPHM_TOC]	+=t;
		current_week		=info.week;

		/* łɓ̂Ȃǂ */
		for(i=0;i<ephm_count[prn-1];i++) {
			/* Tԍv̂Ώ */
			if (ephm_buf[prn-1][i].week!=info.week) continue;

			/* IODCvĂ邩 */
			if (ephm_buf[prn-1][i].data[EPHM_IODC]
					==info.data[EPHM_IODC]) {
				/* M̑̂c */
				if (info.data[EPHM_TOT]
						<ephm_buf[prn-1][i].data[EPHM_TOT]) {
					ephm_buf[prn-1][i]=info;
				}
				prn=0;
				break;
			}
		}
		if (prn<1) continue;

		/* zɊi[iM̏j */
		if (ephm_count[prn-1]>=MAX_EPHMS) {
			fprintf(stderr,"Too long NAV file.\n");
			goto NOERROR;
		}
		for(i=0;i<ephm_count[prn-1];i++) {
			t	=(ephm_buf[prn-1][i].week-info.week)*SECONDS_WEEK
					+ephm_buf[prn-1][i].data[EPHM_TOT];
			if (info.data[EPHM_TOT]<t) break;
		}
		for(j=ephm_count[prn-1];j>i;j--) {
			ephm_buf[prn-1][j]=ephm_buf[prn-1][j-1];
		}
		ephm_buf[prn-1][i]=info;
		ephm_count[prn-1]++;
	}
NOERROR:
	noerr=TRUE;

ERROR:
	/* GtFXȂꍇ */
	if (current_week<0) {
		fprintf(stderr,"Error: No ephemeris information.\n");
		return;
	}

	/* GtFX̂q̐ */
	n=0; for(prn=1;prn<=MAX_PRN;prn++) {
		if (ephm_count[prn-1]>0) n++;
	}
	fprintf(stderr,"week %d: %d satellites\n",current_week,n);

	/* rŃt@CIĂꍇ */
	if (!noerr) {
		fprintf(stderr,"Error: Unexpected EOF.\n");
	}
}

/*------------------------------------------------------------
 * set_ephemeris() - GtFXZbg
 *------------------------------------------------------------
 *  bool set_ephemeris(prn,wt,iode); TRUE:Zbg
 *    int prn;  qPRNԍ(1`)
 *    wtime wt; w
 *    int iode; IODEw/-1:wȂ
 *------------------------------------------------------------*/
bool set_ephemeris(int prn,wtime wt,int iode)
{
	int		i;
	double	t,t0;

	/* VGtFXT */
	for(i=ephm_count[prn-1]-1;i>=0;i--) {
		/* Lł邱 */
		t0	=(ephm_buf[prn-1][i].week-wt.week)*SECONDS_WEEK;
		t	=ephm_buf[prn-1][i].data[EPHM_TOC]+t0;
		if (wt.sec<t-EPHEMERIS_EXPIRE*3600.0-0.1 ||
				t+EPHEMERIS_EXPIRE*3600.0+0.1<wt.sec) continue;

		/* IODE`FbN */
		if (iode>=0) {
			if (ephm_buf[prn-1][i].data[EPHM_IODE]==iode) break;
			continue;
		}

		/* M */
		t	=ephm_buf[prn-1][i].data[EPHM_TOT]+t0;
		if (t<wt.sec+0.1) break;	/* wtO̎ */
	}

	/* JgƂăZbg */
	current_ephm[prn-1]=i;
	return (i>=0);
}

/*------------------------------------------------------------
 * get_ephemeris() - GtFX̃p[^𓾂
 *------------------------------------------------------------
 *  double get_ephemeris(prn,para); p[^l
 *    int prn;  qPRNԍ(1`)
 *    int para; p[^ԍ(0`)
 *------------------------------------------------------------
 *   Oset_ephemeris()ɂGtFXZbgĂ
 * .
 *------------------------------------------------------------*/
double get_ephemeris(int prn,int para)
{
	if (ephm_count[prn-1]<1 || current_ephm[prn-1]<0) {
		fprintf(stderr,"Missing ephemeris: PRN=%d.\n",prn);
		exit(2);
	}

	return ephm_buf[prn-1][current_ephm[prn-1]].data[para];
}

/*------------------------------------------------------------
 * satellite_clock() - qNbN덷vZ
 *------------------------------------------------------------
 *  double satellite_clock(prn,wt,psr); qNbN덷[s]
 *    int prn;    qPRNԍ(1`)
 *    wtime wt;   w
 *    double psr; ̓`[m]
 *------------------------------------------------------------
 *   Oset_ephemeris()ɂGtFXZbgĂ
 * . get_ephemeris()gp.
 *------------------------------------------------------------*/
double satellite_clock(int prn,wtime wt,double psr)
{
	int		i;
	double	tk,tk0,sqrtA,e,n,Ek,Mk,dt,tr;

	/* ԍ߂ */
	tk0	=(wt.week-get_ephemeris(prn,EPHM_WEEK))*SECONDS_WEEK
			+wt.sec-get_ephemeris(prn,EPHM_TOE);
	tk	=tk0-psr/C;							/* `Ԃ̕ */

	/* Sߓ_pEk[rad]߂ */
	sqrtA	=get_ephemeris(prn,EPHM_sqrtA);
	e	=get_ephemeris(prn,EPHM_e); 		/* S */
	n	=sqrt(MUe)/sqrtA/sqrtA/sqrtA+get_ephemeris(prn,EPHM_d_n);
	Mk	=get_ephemeris(prn,EPHM_M0)+n*tk;	/* ϋߓ_p */
	Ek=Mk; for(i=0;i<10;i++) Ek=Mk+e*sin(Ek);	/* Kepler */

	/* Θ_␳ */
	tr	=-2.0*sqrt(MUe)/C/C*e*sqrtA*sin(Ek);

	/* ԍ߂ */
	tk0	=(wt.week-get_ephemeris(prn,EPHM_WEEK))*SECONDS_WEEK
			+wt.sec-get_ephemeris(prn,EPHM_TOC);
	tk	=tk0-psr/C;							/* `Ԃ̕ */

	/* qv̕␳ʂvZ */
	dt	=get_ephemeris(prn,EPHM_AF0)
			+get_ephemeris(prn,EPHM_AF1)*tk
			+get_ephemeris(prn,EPHM_AF2)*tk*tk;

	return dt+tr-get_ephemeris(prn,EPHM_TGD);
}

/*------------------------------------------------------------
 * satellite_position() - qʒuvZ
 *------------------------------------------------------------
 *  posxyz satellite_position(prn,wt,psr); qʒu
 *    int prn;    qPRNԍ(1`)
 *    wtime wt;   w
 *    double psr; ̓`[m]
 *------------------------------------------------------------
 *   Oset_ephemeris()ɂGtFXZbgĂ
 * . get_ephemeris()gp.
 *------------------------------------------------------------*/
posxyz satellite_position(int prn,wtime wt,double psr)
{
	int		i;
	double	tk,tk0,sqrtA,e,n,Ek,Mk,xk,yk,Omegak,
			vk,pk,uk,rk,ik,d_uk,d_rk,d_ik;
	posxyz	pos;

	/* ԍ߂ */
	tk0	=(wt.week-get_ephemeris(prn,EPHM_WEEK))*SECONDS_WEEK
			+wt.sec-get_ephemeris(prn,EPHM_TOE);
	tk	=tk0-psr/C;							/* `Ԃ̕ */

	/* Sߓ_pEk[rad]߂ */
	sqrtA	=get_ephemeris(prn,EPHM_sqrtA);
	e	=get_ephemeris(prn,EPHM_e); 		/* S */
	n	=sqrt(MUe)/sqrtA/sqrtA/sqrtA+get_ephemeris(prn,EPHM_d_n);
	Mk	=get_ephemeris(prn,EPHM_M0)+n*tk;	/* ϋߓ_p */
	Ek=Mk; for(i=0;i<10;i++) Ek=Mk+e*sin(Ek);	/* Kepler */

	/* Oʓɂqʒu */
	rk	=sqrtA*sqrtA*(1.0-e*cos(Ek));		/* a */
	vk	=atan2((sqrt(1.0-e*e)*sin(Ek)),(cos(Ek)-e));/* ^ߓ_p */
	pk	=vk+get_ephemeris(prn,EPHM_omega);	/* ܓx[rad] */

	/* ␳WKp */
	d_uk=get_ephemeris(prn,EPHM_Cus)*sin(2.0*pk)
			+get_ephemeris(prn,EPHM_Cuc)*cos(2.0*pk);
	d_rk=get_ephemeris(prn,EPHM_Crs)*sin(2.0*pk)
			+get_ephemeris(prn,EPHM_Crc)*cos(2.0*pk);
	d_ik=get_ephemeris(prn,EPHM_Cis)*sin(2.0*pk)
			+get_ephemeris(prn,EPHM_Cic)*cos(2.0*pk);
	uk	=pk+d_uk;							/* ܓx[rad] */
	rk	=rk+d_rk;							/* a[m] */
	ik	=get_ephemeris(prn,EPHM_i0)+d_ik
			+get_ephemeris(prn,EPHM_di)*tk;	/* OXΊp[rad] */

	/* Oʓł̈ʒu */
	xk	=rk*cos(uk);
	yk	=rk*sin(uk);

	/* _̌ox[rad] */
	Omegak	=get_ephemeris(prn,EPHM_OMEGA0)
				+(get_ephemeris(prn,EPHM_dOmega)-dOMEGAe)*tk0
				-dOMEGAe*get_ephemeris(prn,EPHM_TOE);

	/* ECEFWnɕϊ */
	pos.x	=xk*cos(Omegak)-yk*sin(Omegak)*cos(ik);
	pos.y	=xk*sin(Omegak)+yk*cos(Omegak)*cos(ik);
	pos.z	=yk*sin(ik);

	return pos;
}

/*------------------------------------------------------------
 * main() - C
 *------------------------------------------------------------*/
void main(int argc,char **argv)
{
	int		n,prn;
	long	epochs,avail,sats;
	double	sec,mask,start,stop,step;
	FILE	*fp;
	wtime	wt;
	posxyz	satpos,usrpos;
	posblh	user;

	/* RINEXq@t@Cǂݍ */
	if (argc<8) {
		fprintf(stderr,"masksats <RINEX-NAV> <lat> <lon> <hgt> \
<start> <stop> <step>\n");
		exit(0);
	} else if ((fp=fopen(argv[1],"rt"))==NULL) {
		perror(argv[1]);
		exit(2);
	} else {
		read_RINEX_NAV(fp);
		fclose(fp);
	}

	/* [Uʒu */
	user.lat=deg_to_rad(atof(argv[2]));
	user.lon=deg_to_rad(atof(argv[3]));
	user.hgt=atof(argv[4]);
	usrpos	=blh_to_xyz(user);

	/* vZΏێ */
	start	=atof(argv[5]);
	stop	=atof(argv[6]);
	step	=atof(argv[7]);

	/* p}XNςȂvZ */
	for(mask=MASK_START;mask-MASK_STOP<0.001;mask+=MASK_STEP) {
		epochs	=0;
		sats	=0;
		avail	=0;

		/* w肳ꂽԔ͈͂ɂČvZ */
		for(sec=start;sec-stop<0.001;sec+=step) {
			wt.week	=current_week;
			wt.sec	=sec;

			/* p\ȉq𐔂 */
			n=0;
			for(prn=1;prn<=MAX_PRN;prn++) {
				/* GtFXZbg */
				if (!set_ephemeris(prn,wt,-1)) continue;

				/* qʒuvZ */
				satpos	=satellite_position(prn,wt,0.0);

				/* p}XNɂI */
				if (elevation(satpos,usrpos)>deg_to_rad(mask)) n++;
			}
			epochs++;
			if (n>=4) avail++;
			sats+=n;
		}

		/* qo */
		printf("%.2f,%d,%.6f,%.6f\n",
			mask,					/* #1: p}XN[deg] */
			epochs,					/* #2: G|bN */
			(double)sats/epochs,	/* #3: ωq */
			(double)avail/epochs);	/* #4: ʉ\G|bN̊ */
	}

	exit(0);
}

